/*
 *  $Id: scatter.c,v 1.4 1993/11/12 12:30:51 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: scatter.c,v 1.4 1993/11/12 12:30:51 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_Scatter - 

@*/
int MPI_Scatter ( sendbuf, sendcnt, sendtype, 
		  recvbuf, recvcnt, recvtype, 
		  root, comm )
void             *sendbuf;
int               sendcnt;
MPI_Datatype      sendtype;
void             *recvbuf;
int               recvcnt;
MPI_Datatype      recvtype;
int               root;
MPI_Comm  comm;
{
  int              rank, size, i, extent;
  MPI_Comm_request req;
  MPI_Status       status;

  /* Get size and rank */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );

  /* Check for invalid arguments */
  if ( (comm            == (MPI_Comm)0) ||
       (root            < 0)                    || 
       (root            >= size)                ||
       (recvcnt         <= 0)                   ||
       (recvbuf         == (void *)0)           ||
       ( (rank == root) && ((sendbuf == (void *)0) || (sendcnt <= 0)) ) ||
       (comm->comm_type == MPIR_INTER) ) 
    return (MPI_SUCCESS); 

  /* Switch communicators to the hidden collective */
  comm = comm->comm_coll;

  /* Get the size of the send type */
  MPI_Extent ( sendtype, &extent );

  /* If I'm the root, send messages to the rest of 'em */
  if ( rank == root ) {
    for ( i=0; i<root; i++ ) {
      MPI_Isend( (void *)((int)sendbuf+i*sendcnt*extent), sendcnt, 
		sendtype, i, MPIR_SCATTER_TAG, comm, &req);
      MPI_Wait ( req, &status );
    }
    MPI_Isend ( (void *)((int)sendbuf+rank*sendcnt*extent),sendcnt,sendtype, 
	        rank, MPIR_SCATTER_TAG, comm, &req );
    MPI_Wait ( req, &status );
    MPI_Irecv ( recvbuf, recvcnt, recvtype, rank, MPIR_SCATTER_TAG, comm, 
	        &req );
    MPI_Wait ( req, &status );
    for ( i=root+1; i<size; i++ ) {
      MPI_Isend( (void *)((int)sendbuf+i*sendcnt*extent), sendcnt, 
		sendtype, i, MPIR_SCATTER_TAG, comm, &req);
      MPI_Wait ( req, &status );
    }
  }
  else {
    MPI_Irecv(recvbuf,recvcnt,recvtype,root,MPIR_SCATTER_TAG,comm,&req);
    MPI_Wait (req, &status );
  }
  return (MPI_SUCCESS);
}
