/*
 *  $Id: u_scan.c,v 1.4 1993/11/12 12:30:51 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: u_scan.c,v 1.4 1993/11/12 12:30:51 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_User_scan - 

@*/
MPI_User_scan ( sendbuf, recvbuf, count, datatype, op, comm )
void             *sendbuf;
void             *recvbuf;
int               count;
MPI_Datatype      datatype;
void            (*op)();
MPI_Comm  comm;
{
  int rank, size, extent;
  MPI_Status status;
  MPI_Comm_request req;

  /* Check for invalid arguments */
  if ( (comm            == (MPI_Comm)0) ||
       (comm->comm_type == MPIR_INTER) ||
       (op              == (void (*)())0) ||
       (count           <= 0) ||
       (sendbuf         == (void *)0) ||
       (recvbuf         == (void *)0) )
    return (MPI_SUCCESS); 
  /* We also need to check that the datatype is a basic type? */

  /* Get my rank & size and switch communicators to the hidden collective */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );
  MPI_Extent ( datatype, &extent );
  comm = comm->comm_coll;

  /* If the recv and send buffers are equivalent */
  if (recvbuf == sendbuf) {
    if (rank > 0) {
      void *buffer = (void *)MALLOC(extent * count);
      MPI_Irecv(buffer,count,datatype,rank-1,MPIR_USER_SCAN_TAG,comm,&req);
      MPI_Wait( req, &status );
      (*op)(buffer, recvbuf, count, datatype); 
      FREE(buffer);
    }
  }
  else  /* send and recv buffers are different */
    if (rank > 0) {
      MPI_Irecv(recvbuf,count,datatype,rank-1,MPIR_USER_SCAN_TAG,comm,&req);
      MPI_Wait( req, &status );
      (*op)(sendbuf, recvbuf, count, datatype); 
    }
    else {
      MPI_Isend(sendbuf,count,datatype,rank,MPIR_USER_SCAN_TAG,comm,&req);
      MPI_Wait( req, &status );
      MPI_Irecv(recvbuf,count,datatype,rank,MPIR_USER_SCAN_TAG,comm,&req);
      MPI_Wait( req, &status );
    }

  /* send the letter to destination */
  if (rank < (size-1)) {
    MPI_Isend(recvbuf,count,datatype,rank+1,MPIR_USER_SCAN_TAG,comm,&req);
    MPI_Wait( req, &status );
  }
  return(MPI_SUCCESS);
}
