/*
 *  $Id: group_excl.c,v 1.3 1993/11/12 12:00:48 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/*@

MPI_Group_excl - 

@*/

#include <mpi.h>


#define MPIR_MARKED 1
#define MPIR_UNMARKED 0

int MPI_Group_excl ( group, n, ranks, group_out )
MPI_Group group, *group_out;
int       n, *ranks;
{
  int i, j, rank;
  MPI_Group new_group;

  /* Check for a NULL input group */
  if ( (group == (MPI_Group)0) || (n >= group->np) ) {
    (*group_out) = (MPI_Group)0;
    return (MPI_SUCCESS);
  }
  
  /* Check for no ranks to exclude */
  if ( n <= 0 ) {
    (void) MPIR_GROUP_dup ( group, group_out );
    return (MPI_SUCCESS);
  }

  /* Create the new group */
  new_group = (*group_out)  = NEW(MPIR_GROUP);
  new_group->ref_count      = 1;
  new_group->local_rank     = MPI_UNDEFINED;
  new_group->set_mark       = (uint32 *)0;
  new_group->np             = group->np - n;
  new_group->lrank_to_grank = (uint32 *)MALLOC(new_group->np * sizeof(uint32));

  /* Allocate set marking space for group if necessary */
  if (group->set_mark == NULL)
    group->set_mark = (uint32 *) MALLOC( group->np * sizeof(uint32) );
  (void) memset( group->set_mark, (char)0, group->np * sizeof(uint32) );

  /* Mark the ranks to be excluded */
  for (i=0; i<n; i++) 
    if ( ((rank = ranks[i]) < group->np) && (rank >= 0) ) 
      group->set_mark[rank] = MPIR_MARKED;

  /* Fill in the lrank_to_grank list */
  for (i=j=0; i < group->np ; i++) 
    if ( (group->set_mark[i] == MPIR_UNMARKED) && (j < new_group->np ) ) {
      if (group->local_rank == i)
        new_group->local_rank = j;
      new_group->lrank_to_grank[j++] = group->lrank_to_grank[i];
    }

  /* Determine the previous and next powers of 2 */
  MPIR_Powers_of_2 ( new_group->np, &(new_group->N2_next), &(new_group->N2_prev) );

  return (MPI_SUCCESS);
}
